import copy
import json
import os

import torch
from torch_geometric.data import Data
from torch_geometric.explain.explanation import Explanation


def get_explanation(explainer, item):
    explanation = explainer(
        x=item.x,
        edge_index=item.edge_index,
        index=int(item.y),
        target=item.y,
    )
    # TODO return None if pas bien plutot
    assert explanation_verification(explanation)
    return explanation


def is_empty_graph(data: Data) -> bool:
    return data.x.shape[0] == 0


def get_pred(explainer, explanation):
    pred = explainer.get_prediction(x=explanation.x, edge_index=explanation.edge_index)[
        0
    ]
    setattr(explanation, "pred", pred)
    data = explanation.to_dict()
    if not data.get("node_mask") is None or not data.get("edge_mask") is None:
        pred_masked = explainer.get_masked_prediction(
            x=explanation.x,
            edge_index=explanation.edge_index,
            node_mask=data.get("node_mask"),
            edge_mask=data.get("edge_mask"),
        )[0]
        setattr(explanation, "pred_exp", pred_masked)


def explanation_verification(exp: Explanation) -> bool:
    is_good = True
    masks = [v for k, v in exp.items() if "_mask" in k and isinstance(v, torch.Tensor)]
    for mask in masks:
        is_nan = mask.isnan().any().item()
        is_inf = mask.isinf().any().item()
        is_const = mask.max() == mask.min()
        is_ok = exp.validate()
        if is_nan or is_inf or not is_ok or is_const:
            is_good = False
            return is_good
        else:
            continue
    return is_good


def save_explanation(exp: Explanation, path: str) -> None:
    data = copy.copy(exp).to_dict()
    for k, v in data.items():
        if isinstance(v, torch.Tensor):
            data[k] = v.detach().cpu().tolist()

    with open(path, "w") as f:
        json.dump(data, f)


def load_explanation(path: str) -> Explanation:
    with open(path, "r") as f:
        data = json.load(f)
    for k, v in data.items():
        if isinstance(v, list):
            if k == "edge_index" or k == "y":
                data[k] = torch.LongTensor(v)
            else:
                data[k] = torch.FloatTensor(v)
    return Explanation.from_dict(data)


def normalize_explanation_masks(exp: Explanation, p: str = "inf") -> Explanation:
    exp = copy.copy(exp)
    data = exp.to_dict()
    for k, v in data.items():
        if "_mask" in k and isinstance(v, torch.FloatTensor):
            norm = torch.norm(input=data[k], p=p, dim=None).item()
            if norm.item() > 0:
                data[k] = data[k] / norm
    return exp
